from datetime import datetime
from typing import Literal

from pydantic import BaseModel, Field


class DataModel(BaseModel):
    name: str
    schema_field: str = Field(..., alias="schema")
    database: str
    columns: list[str]


class DashboardOwner(BaseModel):
    user_id: str  #  Typically an email or a uuid.
    username: (
        str | None
    )  # Typically an email address. Empty in PBI if a group or app is the owner.
    name: str  # Typically the human name.


class Dashboard(BaseModel):
    external_id: str
    url: str
    type: Literal["powerbi", "qliksense"]
    name: str
    workspace: str
    created_at: datetime | None
    modified_at: datetime
    owners: list[DashboardOwner]
    parent_models: list[DataModel]


class DashboardSchemaMetadata(BaseModel):
    schema_field: str = Field(..., alias="schema")
    version: int


class DashboardManifest(BaseModel):
    metadata: DashboardSchemaMetadata
    payload: list[Dashboard]



import socket
from pathlib import Path
from time import sleep

import multiprocess
import psycopg2
import pytest
import uvicorn
from fastapi import FastAPI, HTTPException
from pytest_postgresql import factories

app = FastAPI()


@app.post("/api/v1/dbt")
async def dbt(body: dict):
    if body is None:
        raise HTTPException(400)
    return body


@app.post("/api/v1/dbt/run_results")
async def run_results(body: dict):
    if body is None:
        raise HTTPException(400)
    return body


@app.post("/api/v1/postgres")
async def postgres(body: dict):
    if body is None:
        raise HTTPException(400)
    return body


@app.post("/api/v1/config")
async def config(body: dict):
    if body is None:
        raise HTTPException(400)
    return body


def find_free_port():
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("", 0))
        return s.getsockname()[1]


@pytest.fixture(scope="module")
def test_server():
    port = find_free_port()
    server = multiprocess.Process(
        target=uvicorn.run,
        args=(app,),
        kwargs={"host": "0.0.0.0", "port": port},
        daemon=True,
    )
    server.start()
    sleep(1)  # Time to wait while the server starts up
    yield f"http://0.0.0.0:{port}"
    server.terminate()


@pytest.fixture(scope="module")
def config_dir(LUMA_CONFIG_DIR):
    config_dir = Path(LUMA_CONFIG_DIR)
    config_path = config_dir / "config.yaml"
    owners_path = config_dir / "owners.yaml"

    config_dir.mkdir(exist_ok=True)
    config_path.touch()
    with config_path.open("w") as f:
        f.write(
            """groups:
  - meta_key: "domain"
    slug: "domains"
    label_plural: "Domains"
    label_singular: "Domain"
    icon: "Cube"
    in_sidebar: true
    visible: true
  - meta_key: "true_source"
    slug: "sources"
    label_plural: "Sources"
    label_singular: "Source"
    icon: "Cloud"
    in_sidebar: false
"""
        )
    owners_path.touch()
    with owners_path.open("w") as f:
        f.write(
            """owners:
- email: "some@one.com"
  first_name: "Dave"
  last_name: "Smith"
  title: "Director"
- email: "other@person.com"
  first_name: "Michelle"
  last_name: "Dunne"
  title: "CTO"
- email: "someone@else.com"
  first_name: "Dana"
  last_name: "Pawlak"
  title: "HR Manager"
"""
        )
    yield config_dir

    config_path.unlink()
    owners_path.unlink()
    config_dir.rmdir()


@pytest.fixture(scope="module")
def config_dir_invalid_schema(INVALID_SCHEMA_LUMA_CONFIG_DIR):
    config_dir = Path(INVALID_SCHEMA_LUMA_CONFIG_DIR)
    config_path = config_dir / "config.yaml"
    owners_path = config_dir / "owners.yaml"

    config_dir.mkdir(exist_ok=True)
    config_path.touch()
    with config_path.open("w") as f:
        f.write(
            """groups:
  - meta_key: ["domain"]
    slug: ["domains"]
    label_plural: "Domains"
    label_singular: "Domain"
    icon: "Cube"
  - meta_key: ["true_source"]
    slug: "sources"
    label_plural: "Sources"
    label_singular: "Source"
    icon: "Cloud"
"""
        )
    owners_path.touch()
    with config_path.open("w") as f:
        f.write(
            """owners:
- email: "some@one.com"
  first_name: "Dave"
  last_name: "Smith"
  title: "Director"
- email: "Director.com"
  first_name: "Michelle"
  last_name: "Dunne"
  title: "CTO"
- email:["someone@else.com"]
  first_name: "Dana"
  last_name: "Pawlak"
  title: "HR Manager"
"""
        )
    yield config_dir
    config_path.unlink()
    owners_path.unlink()
    config_dir.rmdir()


# This fixture creates a Postgres database
postgresql_proc = factories.postgresql_proc(port=find_free_port())
postgresql = factories.postgresql("postgresql_proc")


@pytest.fixture(scope="function")
def setup_db(postgresql):
    # Get the connection details from the postgresql fixture
    user = postgresql.info.user
    password = postgresql.info.password
    dbname = postgresql.info.dbname
    host = postgresql.info.host
    port = postgresql.info.port

    conn = psycopg2.connect(
        dbname=dbname, user=user, password=password, host=host, port=port
    )
    cur = conn.cursor()

    # Creating tables and inserting test data
    cur.execute(
        """
        CREATE TABLE users (
            id SERIAL PRIMARY KEY,
            name VARCHAR(50),
            email VARCHAR(50),
            created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP
        );
        INSERT INTO users (name, email) VALUES
            ('John Doe', 'john@example.com'),
            ('Jane Smith', 'jane@example.com'),
            ('Mike Johnson', 'mike@example.com');
        
        CREATE TABLE products (
            id SERIAL PRIMARY KEY,
            name VARCHAR(100),
            price DECIMAL(10, 2),
            created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP
        );
        INSERT INTO products (name, price) VALUES
            ('Product A', 19.99),
            ('Product B', 9.99),
            ('Product C', 14.99);
        
        CREATE TABLE orders (
            id SERIAL PRIMARY KEY,
            user_id INTEGER REFERENCES users (id),
            product_id INTEGER REFERENCES products (id),
            quantity INTEGER,
            created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP
        );
        INSERT INTO orders (user_id, product_id, quantity) VALUES
            (1, 1, 2),
            (2, 2, 1),
            (3, 3, 3);

        CREATE OR REPLACE VIEW user_orders AS
        SELECT
            users.name AS user_name,
            users.email AS user_email,
            products.name AS product_name,
            orders.quantity AS order_quantity,
            orders.created_at AS order_date
        FROM
            users
        JOIN
            orders ON users.id = orders.user_id
        JOIN
            products ON products.id = orders.product_id;

        CREATE OR REPLACE VIEW product_sales AS
        SELECT
            products.name AS product_name,
            SUM(orders.quantity) AS total_quantity_sold,
            SUM(orders.quantity * products.price) AS total_revenue
        FROM
            products
        JOIN
            orders ON products.id = orders.product_id
        GROUP BY
            products.name;
    """
    )

    connection.commit()
    cursor.close()

    yield connection

    cursor = connection.cursor()
    cursor.execute("DROP SCHEMA test_schema;")
    connection.commit()
    connection.close()


@pytest.fixture(scope="module")
def TESTS_DIR():
    return Path(__file__).parent

@pytest.fixture(scope="module")
def METADATA_DIR_V1_7(TESTS_DIR):
    return TESTS_DIR / "assets" / "dbt_v1.7"


@pytest.fixture(scope="module")
def METADATA_DIR_V1_9(TESTS_DIR):
    return TESTS_DIR / "assets" / "dbt_v1.9"


@pytest.fixture(scope="module")
def INVALID_SCHEMA_LUMA_CONFIG_DIR(TESTS_DIR):
    return TESTS_DIR / ".invalid_schema_luma"


@pytest.fixture(scope="module")
def LUMA_CONFIG_DIR(TESTS_DIR):
    return TESTS_DIR / ".luma"



from pydantic import BaseModel, Field


class DatabaseTableColumn(BaseModel):
    name: str
    type: str
    length: str
    description: str | None


class DatabaseTable(BaseModel):
    type: str
    name: str
    columns: list[DatabaseTableColumn]


class DatabaseTableSchemaMetadata(BaseModel):
    schema_field: str = Field(..., alias="schema")
    version: int


class DatabaseTableManifest(BaseModel):
    metadata: DatabaseTableSchemaMetadata
    payload: list[DatabaseTable]



"""Ingest dbt metadata into Luma."""

from enum import Enum
from pathlib import Path
import time

import typer
from dbt_artifacts_parser.parser import (
    parse_catalog,
    parse_manifest,
    parse_run_results,
)
from dbt_artifacts_parser.parsers.utils import get_dbt_schema_version
from rich import print
from rich.console import Console
from rich.panel import Panel
import typer

from lumaCLI.utils import (
    check_ingestion_results,
    check_ingestion_status, perform_ingestion_request,
)
from lumaCLI.utils.options import (
    DryRun,
    Follow,
    LumaURL,
    MetadataDir,
)

from lumaCLI.utils.luma_utils import json_path_to_dict

app = typer.Typer(no_args_is_help=True, pretty_exceptions_show_locals=False)
console = Console()


class IngestionStatus(Enum):
    successful = 0
    failed = 1
    pending = 2


# A matrix of supported dbt versions and their corresponding artifact schema versions.
# This provides a clear source of truth for supported versions.
SUPPORTED_DBT_VERSIONS = {
    "https://schemas.getdbt.com/dbt/manifest/v7.json",
    "https://schemas.getdbt.com/dbt/manifest/v11.json",
    "https://schemas.getdbt.com/dbt/manifest/v12.json",
    "https://schemas.getdbt.com/dbt/catalog/v1.json",
    "https://schemas.getdbt.com/dbt/run-results/v4.json",
    "https://schemas.getdbt.com/dbt/run-results/v5.json",
    "https://schemas.getdbt.com/dbt/run-results/v6.json",
}

def _is_supported_dbt_version(dbt_version: str) -> bool:
    """
    Check if the provided dbt version is supported.
    """
    return dbt_version in SUPPORTED_DBT_VERSIONS

def _validate_dbt_version(artifact_dict: dict):
    """
    Check if the dbt version in the artifact dictionary is supported.
    Raises an exception if the version is not supported.
    """
    dbt_version = get_dbt_schema_version(artifact_dict)
    if not _is_supported_dbt_version(dbt_version):
        raise Exception(
            f"Unsupported dbt version: {dbt_version}. Supported versions are: {', '.join(SUPPORTED_DBT_VERSIONS.keys())}"
        )

@app.command()
def validate(
    metadata_dir: Path = MetadataDir,
):
    """
    Validates dbt artifacts (manifest.json, catalog.json, etc.) using dbt-artifacts-parser.
    """
    all_valid = True
    found_any = False
    console.print(
        Panel(f"🔍 Validating artifacts in [cyan]{metadata_dir.absolute()}[/cyan]/")
    )

    def _validate_artifact(file_path, parser_func, artifact_name):
        nonlocal all_valid, found_any
        if not file_path.is_file():
            return

        found_any = True
        try:
            artifact_dict = json_path_to_dict(file_path)
            dbt_version = get_dbt_schema_version(artifact_dict)
            if not _is_supported_dbt_version(dbt_version):
                raise Exception(f"Unsupported dbt version: {dbt_version}"
                    f"Unsupported dbt version: {dbt_version}. Supported versions are: {', '.join(SUPPORTED_DBT_VERSIONS.keys())}"
                )
            parser_func(artifact_dict)
            console.print(f"[green]✔ {artifact_name} is valid.[/green]")
        except Exception as e:
            console.print_exception(show_locals=True)
            console.print(
                Panel(f"[red]{e}[/red]", title=f"Error processing: {artifact_name}:", border_style="red")
            )
            # add traceback to the console output
            console.print(
                Panel(
                    f"[red]Error processing {artifact_name} at {file_path.absolute()}[/red]",
                    border_style="red",
                )
            )
            all_valid = False

    artifacts_to_check = [
        (metadata_dir / "manifest.json", parse_manifest, "manifest.json"),
        (metadata_dir / "catalog.json", parse_catalog, "catalog.json"),
    ]

    for path, artifact_parser, name in artifacts_to_check:
        _validate_artifact(path, artifact_parser, name)

    if not found_any:
        console.print(
            Panel("[bold yellow]No dbt artifacts found to validate.[/bold yellow]")
        )
        raise typer.Exit(1)

    if all_valid:
        console.print(
            Panel("[bold green]✅ All found artifacts are valid![/bold green]")
        )
    else:
        console.print(
            Panel("[bold red]❌ Validation failed for one or more artifacts.[/bold red]")
        )
        raise typer.Exit(1)


@app.command()
def ingest(  # noqa: C901
    metadata_dir: Path = MetadataDir,
    luma_url: str = LumaURL,
    dry_run: bool = DryRun,
    follow: bool = Follow,
):
    """
    Validates and ingests a bundle of dbt artifacts from the specified directory.
    manifest.json and catalog.json are required.
    """
    # Define JSON paths.
    manifest_json_path = metadata_dir / "manifest.json"
    catalog_json_path = metadata_dir / "catalog.json"

    # Ensure required files exist.
    if not manifest_json_path.is_file():
        console.print(
            Panel(f"[red]Required file not found: {manifest_json_path.absolute()}[/red]")
        )
    if not catalog_json_path.is_file():
        console.print(Panel(f"[red]Required file not found: {catalog_json_path.absolute()}[/red]"))
        raise typer.Exit(1)

    console.print(
        Panel(
            f"📦 Preparing dbt artifacts from [cyan]{metadata_dir.absolute()}[/cyan] for ingestion..."
        )
    )

    # Validate and load artifacts
    try:
        manifest_dict = json_path_to_dict(manifest_json_path)
        _validate_dbt_version(manifest_dict)
        manifest_artifact = parse_manifest(manifest_dict)
        console.print("[green]✔ manifest.json validated and loaded.[/green]")

        catalog_dict = json_path_to_dict(catalog_json_path)
        _validate_dbt_version(catalog_dict)
        catalog_artifact = parse_catalog(catalog_dict)
        console.print("[green]✔ catalog.json validated and loaded.[/green]")

    except Exception as e:
        console.print(
            Panel(f"[red]✖ Artifact validation failed:[/red]\n{e}", border_style="red")
        )
        raise e

    # Define bundle dict
    artifacts_bundle = {
        "manifest_json": manifest_artifact.model_dump(by_alias=True, mode='json'),
        "catalog_json": catalog_artifact.model_dump(by_alias=True, mode='json'),
    }

    # If in dry run mode, print the bundle and exit.
    if dry_run:
        print("Dry run mode: Payload with the following keys would be sent:")
        print(list(artifacts_bundle.keys()))
        raise typer.Exit(0)

    # Send ingestion request.
    endpoint = f"{luma_url}/api/v1/dbt/"
    response, ingestion_uuid = perform_ingestion_request(
        url=endpoint,
        method="POST",
        payload=artifacts_bundle,
        verify=False,
    )
    if not response.ok:
        raise typer.Exit(1)

    # Wait until ingestion is complete.
    if follow and ingestion_uuid:
        ingestion_status = None

        with console.status("Waiting...", spinner="dots"):
            for _ in range(30):
                ingestion_status = check_ingestion_status(
                    luma_url, ingestion_uuid, verify=False
                )
                if ingestion_status == IngestionStatus.successful.value:
                    response = check_ingestion_results(luma_url, ingestion_uuid)
                    print()
                    print(f"Ingestion results for ID {ingestion_uuid}:")
                    print()
                    print(response)
                    return

                if ingestion_status == IngestionStatus.failed.value:
                    print()
                    print(f"Ingestion failed for ID {ingestion_uuid}")
                    return

                if ingestion_status == IngestionStatus.pending.value:
                    time.sleep(1)

        if ingestion_status != IngestionStatus.successful.value:
            print(
                f"Ingestion did not complete successfully within the wait period. Status: {ingestion_status}"
            )


@app.command()
def send_test_results(
    metadata_dir: Path = MetadataDir,
    luma_url: str = LumaURL,
    dry_run: bool = DryRun,
    follow: bool = Follow,
):
    """
    Validates and sends 'run_results.json' from the specified directory to Luma.
    """

    # Define the path to 'run_results.json'
    run_results_path = metadata_dir / "run_results.json"

    if not run_results_path.is_file():
        print(Panel(f"[red]File not found: {run_results_path.absolute()}[/red]"))
        raise typer.Exit(1)

    console.print(
        Panel(
            f"📦 Preparing run_results.json from [cyan]{metadata_dir.absolute()}[/cyan]..."
        )
    )

    # Validate and load artifact
    try:
        run_results_dict = json_path_to_dict(run_results_path)
        run_results_artifact = parse_run_results(run_results_dict).model_dump(by_alias=True, mode='json')

        console.print("[green]✔ run_results.json validated and loaded.[/green]")
    except Exception as e:
        console.print(
            Panel(f"[bold red]✖ Artifact validation failed:[/bold red]\n{e}", border_style="red")
        )
        raise typer.Exit(1)

    # If in dry run mode, print the test results and exit.
    if dry_run:
        print("Dry run mode: The following payload would be sent:")
        print(run_results_dict)
        raise typer.Exit(0)

    # Send ingestion request.
    endpoint = f"{luma_url}/api/v1/dbt/run_results/"
    response, ingestion_uuid = perform_ingestion_request(
        url=endpoint,
        method="POST",
        payload=run_results_artifact,
        verify=False,
    )
    if not response.ok:
        raise typer.Exit(1)

    if follow and ingestion_uuid:
        ingestion_status = None

        with console.status("Waiting...", spinner="dots"):
            for _ in range(30):
                ingestion_status = check_ingestion_status(luma_url, ingestion_uuid)
                if ingestion_status == IngestionStatus.successful.value:
                    response = check_ingestion_results(luma_url, ingestion_uuid)
                    print()
                    print(f"Ingestion results for ID {ingestion_uuid}:")
                    print()
                    print(response)
                    return

                if ingestion_status == IngestionStatus.failed.value:
                    print()
                    print(f"Ingestion failed for ID {ingestion_uuid}")
                    return

                if ingestion_status == IngestionStatus.pending.value:
                    time.sleep(1)

        if ingestion_status != IngestionStatus.successful.value:
            print(
                f"Ingestion did not complete successfully within the wait period. Status: {ingestion_status}"
            )


# Run the application.
if __name__ == "__main__":
    app()


from lumaCLI.metadata.sources.powerbi.extract import powerbi
from lumaCLI.metadata.sources.powerbi.models import WorkspaceInfo
from lumaCLI.metadata.sources.powerbi.transform import transform


def pipeline():
    source = powerbi()
    metadata = WorkspaceInfo(**next(iter(source)))
    yield transform(metadata)


if __name__ == "__main__":
    from pathlib import Path

    manifest = next(iter(pipeline()))

    with Path("powerbi_extracted.json").open("w", encoding="utf-8") as f:
        f.write(manifest.model_dump_json(by_alias=True))
    # print(manifest)



from lumaCLI.metadata.sources.powerbi.transform import (
    _extract_dataflow_table_from_expression,
    _extract_table_from_expression,
)


def test__extract_table_from_expression():
    expr = 'let\n    Source = AmazonRedshift.Database("redshift-cluster.someaddr123.eu-west-1.redshift.amazonaws.com","somedb"),\n    sandbox = Source{[Name="finance"]}[Data],\n    kpi_p1 = sandbox{[Name="fi_kpi_p1_table"]}[Data]\nin\n    kpi_p1'
    expected = {
        "name": "fi_kpi_p1_table",
        "schema": "finance",
        "database": "somedb",
    }
    assert _extract_table_from_expression(expr) == expected

    expr2 = 'let\n    Source_Redshift = AmazonRedshift.Database("redshift-cluster.someaddr123.eu-west-1.redshift.amazonaws.com","somedb"),\n    finance = Source_Redshift{[Name="finance"]}[Data],\n    fi_zsd_familt = finance{[Name="fi_zsd_familt"]}[Data],\n    #"Renamed Columns" = Table.RenameColumns(fi_zsd_familt,{{"sub_family_code", "Family Code"}, {"sub_family_desc", "Family"}})\nin\n    #"Renamed Columns"'
    expected2 = {
        "name": "fi_zsd_familt",
        "schema": "finance",
        "database": "somedb",
    }
    assert _extract_table_from_expression(expr2) == expected2

    expr3 = 'let\n    Source_Redshift = AmazonRedshift.Database("redshift-cluster.someaddr123.eu-west-1.redshift.amazonaws.com","somedb"),\n    finance = Source_Redshift{[Name="finance"]}[Data],\n    fi_tcurr = finance{[Name="fi_tcurr"]}[Data],\n    #"Renamed Columns2" = Table.RenameColumns(fi_tcurr,{{"exch_rate_type", "Exch Rate Type"}, {"from_currency", "From Currency"}, {"to_currency", "To Currency"}, {"valid_from", "Valid From"}, {"ratio_from", "Ratio From"}, {"exchange_rate", "Exchange Rate"}, {"ratio_to", "Ratio To"}}),\n    #"Changed Type Valid From" = Table.TransformColumnTypes(#"Renamed Columns2", {{"Valid From", type date}}, "es-ES"),\n    \n    #"Filtered BUDG & M" = Table.SelectRows(#"Changed Type Valid From", each ([Exch Rate Type] = "BUDG" or [Exch Rate Type] = "M") ),\n    #"Filtered From Curr EUR" = Table.SelectRows(#"Filtered BUDG & M", each ([From Currency] = "EUR") ),\n    #"Add Year" = Table.AddColumn(#"Filtered From Curr EUR", "Year", each Date.Year([Valid From])),\n    #"Add Month" = Table.AddColumn(#"Add Year", "Month", each Date.Month([Valid From])),\n    #"Add Day" = Table.AddColumn(#"Add Month", "Day", each Date.Day([Valid From])),\n    #"Changed Data Types" = Table.TransformColumnTypes(#"Add Day",{{"Year", Int64.Type}, {"Month", Int64.Type}, {"Day", Int64.Type}}),\n    #"Filtered Day 1" = Table.SelectRows(#"Changed Data Types", each ([Day] = 1) ),\n    #"Renamed Columns" = Table.RenameColumns(#"Filtered Day 1",{{"Valid From", "EX Date"}}),\n    #"Filtered Rows" = Table.SelectRows(#"Renamed Columns", each [Year] >= 2022)\nin\n    #"Filtered Rows"'
    expected3 = {
        "name": "fi_tcurr",
        "schema": "finance",
        "database": "somedb",
    }
    assert _extract_table_from_expression(expr3) == expected3

    # NativeQuery example - we don't extract models from those, as they're unstructured.
    expr4 = "let\n    Source_Redshift = Value.NativeQuery(AmazonRedshift.Database(\"redshift-cluster.someaddr123.eu-west-1.redshift.amazonaws.com\",\"somedb\"), \"SELECT #(lf)  LOWER(p.model) AS model,#(lf)  s.product_line,#(lf)  s.product_line_detail,#(lf)  SUM(COALESCE(s.placements, 0)) AS total_placements#(lf)FROM finance.fi_part_numbers_for_placements p#(lf)LEFT JOIN (#(lf)  SELECT#(lf)    material,#(lf)    product_line,#(lf)    product_line_detail,#(lf)    COALESCE(SUM(sales_qty), 0) + COALESCE(SUM(actual_qty_reagent_rental), 0) AS placements#(lf)  FROM finance.fi_zsd_sald#(lf)  WHERE fiscal_year BETWEEN EXTRACT(YEAR FROM CURRENT_DATE) - 2 #(lf)                         AND EXTRACT(YEAR FROM CURRENT_DATE) + 1#(lf)    AND sales_org IN (#(lf)      '001', '005', '006', '015', #(lf)      '133', '134', '135', #(lf)      '1693', '175', '176', #(lf)      '193', '1933'#(lf)    )#(lf)  GROUP BY material, product_line, product_line_detail#(lf)) s ON p.material = s.material#(lf)GROUP BY LOWER(p.model), s.product_line, s.product_line_detail#(lf)HAVING SUM(COALESCE(s.placements, 0)) <> 0\", null, [EnableFolding=true])\nin\n    Source_Redshift"
    assert _extract_table_from_expression(expr4) is None


def test__extract_dataflow_table_from_expression():
    expr = 'let\n    Source = SourceNewStrategicDim,\n   Table = Source{[entity="co_sales_org"]}[Data],\n   Text = Table.TransformColumnTypes(Table, List.Transform(Table.ColumnNames(Table),each {_, type text} )),\n    Trim = Table.TransformColumns(Text,List.Transform(Table.ColumnNames(Text), each {_, each Text.Trim(_), type text})),\n    Null =  Table.ReplaceValue(Trim,"",null,Replacer.ReplaceValue,Table.ColumnNames(Trim)),\n    Proper = Table.TransformColumnNames(Null, Text.Proper )\nin\n    Proper'
    expected = {
        "name": "co_sales_org",
        "schema": None,
        "database": None,
    }
    assert _extract_dataflow_table_from_expression(expr) == expected



from unittest.mock import patch
import os
from pathlib import Path

import pytest
from typer.testing import CliRunner

from lumaCLI.commands.dbt import validate, IngestionStatus
from lumaCLI.luma import app

runner = CliRunner()

METADATA_PATH_FIXTURES = ["METADATA_DIR_V1_7", "METADATA_DIR_V1_9"]

SUCCESS = 0
FAILURE = 1

@pytest.mark.parametrize("metadata_dir_fixture", METADATA_PATH_FIXTURES)
def test_ingest_requires_only_manifest_and_catalog_json(
    test_server, request, metadata_dir_fixture
):
    """
    Test if 'ingest' requires only manifest.json and catalog.json.
    Parameterized to run against multiple dbt versions.
    """
    # Use the 'request' fixture to get the actual value of the fixture
    # whose name is passed as a string (e.g., "METADATA_DIR_V1_7")
    metadata_dir = f"{request.getfixturevalue(metadata_dir_fixture)}/model_metadata"

    # Get result when the required files are present
    result_no_changes = _get_result(test_server, "ingest", metadata_dir=metadata_dir)

    # Get result when the rest of the json files are missing
    result_manifest_missing = _get_result_whilst_file_missing(
        test_server, "ingest", "manifest.json", metadata_dir=metadata_dir
    )
    result_catalog_missing = _get_result_whilst_file_missing(
        test_server, "ingest", "catalog.json", metadata_dir=metadata_dir
    )

    # Check ingest works when no files are changed, or sources.json or run_results.json are missing
    assert result_no_changes.exit_code == SUCCESS
    assert result_manifest_missing.exit_code == FAILURE
    assert result_catalog_missing.exit_code == FAILURE


@pytest.mark.parametrize("metadata_dir_fixture", METADATA_PATH_FIXTURES)
def test_send_test_results_with_shorthand(
    test_server, request, metadata_dir_fixture
):
    """
    Test if 'send-test-results' with shorthand '-m' requires only run_results.json.
    Parameterized to run against multiple dbt versions.
    """
    # Use the 'request' fixture to get the actual value of the fixture
    metadata_dir = f"{request.getfixturevalue(metadata_dir_fixture)}/model_run_metadata"

    # Get result when the required files are present
    result = _get_result_with_shorthand(
        test_server, "send-test-results", metadata_dir=metadata_dir
    )

    assert result.exit_code == SUCCESS


@pytest.mark.parametrize("metadata_dir_fixture", METADATA_PATH_FIXTURES)
def test_send_test_results_with_shorthand(
    test_server, request, metadata_dir_fixture
):
    """
    Test if 'send-test-results' with shorthand '-m' requires only run_results.json.
    Parameterized to run against multiple dbt versions.
    """
    # Use the 'request' fixture to get the actual value of the fixture
    metadata_dir = f"{request.getfixturevalue(metadata_dir_fixture)}/model_run_metadata"

    # Get result when the required files are present
    result = _get_result_with_shorthand(
        test_server, "send-test-results", metadata_dir=metadata_dir
    )

    assert result.exit_code == SUCCESS


@pytest.mark.parametrize("metadata_dir_fixture", METADATA_PATH_FIXTURES)
def test_send_test_results(
    test_server, request, metadata_dir_fixture
):
    """
    Test if 'send-test-results' requires only run_results.json.
    Parameterized to run against multiple dbt versions.
    """
    # Use the 'request' fixture to get the actual value of the fixture
    metadata_dir = f"{request.getfixturevalue(metadata_dir_fixture)}/model_run_metadata"

    # Get result when the required files are present
    result = _get_result(
        test_server, "send-test-results", metadata_dir=metadata_dir
    )

    assert result.exit_code == SUCCESS


@pytest.mark.parametrize("metadata_dir_fixture", METADATA_PATH_FIXTURES)
def test_send_test_results_dry_run_with_shorthand(
    test_server, request, metadata_dir_fixture
):
    """
    Test if 'send-test-results' with shorthand '-D' correctly performs a dry run.
    Parameterized to run against multiple dbt versions.
    """
    # Use the 'request' fixture to get the actual value of the fixture
    metadata_dir = f"{request.getfixturevalue(metadata_dir_fixture)}/model_run_metadata"

    # Get result when the required files are present
    result = runner.invoke(
        app,
        [
            "dbt",
            "send-test-results",
            "-m",
            metadata_dir,
            "-l",
            test_server,
            "-D",
        ],
    )

    assert result.exit_code == SUCCESS
    assert "Dry run mode" in result.stdout





@pytest.mark.parametrize("metadata_dir_fixture", METADATA_PATH_FIXTURES)
def test_dbt_validate_on_valid_files(test_server, request, metadata_dir_fixture):
    """
    Test if 'dbt validate' works with the provided metadata directory.
    Parameterized to run against multiple dbt versions.
    """
    metadata_dir = f"{request.getfixturevalue(metadata_dir_fixture)}/model_metadata"

    try:
        validate(metadata_dir=Path(metadata_dir))
    except Exception as exc:
        assert False, f"validate raised an exception {exc}"

def test_dbt_validate_on_missing_files():
    """
    Test if 'dbt validate' fails on non-existing paths.
    Parameterized to run against multiple dbt versions.
    """
    metadata_dir = f"dummy"

    with pytest.raises(Exception):
        validate(metadata_dir=Path(metadata_dir))


def _get_result_whilst_file_missing(test_server, command, file_name, metadata_dir):
    """Rename a file, run the command, and rename it back."""
    original_path = os.path.join(metadata_dir, file_name)
    renamed_path = os.path.join(metadata_dir, f"missing_{file_name}")

    os.rename(original_path, renamed_path)
    try:
        result = _get_result(test_server, command, metadata_dir=metadata_dir)
    finally:
        os.rename(renamed_path, original_path)
    return result


def _get_result(test_server, command, metadata_dir):
    """Invoke the CLI runner and return the result."""
    result = runner.invoke(
        app,
        [
            "dbt",
            command,
            "--metadata-dir",
            metadata_dir,
            "--luma-url",
            test_server,
        ],
    )
    return result


def _get_result_with_shorthand(test_server, command, metadata_dir):
    """Invoke the CLI runner with shorthand and return the result."""
    result = runner.invoke(
        app,
        [
            "dbt",
            command,
            "-m",
            metadata_dir,
            "-l",
            test_server,
        ],
    )
    return result





"""Download lineage information from PowerBI API."""

from collections.abc import Generator
from pathlib import Path
from time import sleep
from typing import Any

from azure.identity import ClientSecretCredential
import dlt
from dlt.common import logger
from dlt.sources.helpers.rest_client import RESTClient
from dlt.sources.helpers.rest_client.auth import OAuth2AuthBase
from dlt.sources.helpers.rest_client.paginators import SinglePagePaginator


Workspace = dict[str, Any]


class PowerBIOauthClientCredentials(OAuth2AuthBase):
    def __init__(self, tenant_id: str, client_id: str, client_secret: str):
        """PowerBI OAuth2 client credentials authentication.

        Args:
            tenant_id (str): The Azure tenant ID.
            client_id (str): The client ID of the service principal app.
            client_secret (str): The client secret of the service principal app.
        """
        super().__init__()
        self.access_token = self._get_token(tenant_id, client_id, client_secret)

    @staticmethod
    def _get_token(tenant_id: str, client_id: str, client_secret: str) -> str:
        scope = "https://analysis.windows.net/powerbi/api/.default"
        client_secret_credential_class = ClientSecretCredential(
            tenant_id=tenant_id, client_id=client_id, client_secret=client_secret
        )
        return client_secret_credential_class.get_token(scope).token


@dlt.source(name="powerbi")
def powerbi(
    tenant_id: str = dlt.secrets.value,
    client_id: str = dlt.secrets.value,
    client_secret: str = dlt.secrets.value,
) -> dict:
    """The PowerBI metadata source.

    Args:
        tenant_id (str, optional): The Azure tenant ID. Defaults to dlt.secrets.value.
        client_id (str, optional): The client ID of the service principal app. Defaults
            to dlt.secrets.value.
        client_secret (str, optional): The client secret of the service principal app.
            Defaults to dlt.secrets.value.

    Returns:
        dict: _description_

    Yields:
        Iterator[dict]: _description_
    """
    client = RESTClient(
        base_url="https://api.powerbi.com/v1.0/myorg",
        auth=PowerBIOauthClientCredentials(
            tenant_id=tenant_id,
            client_id=client_id,
            client_secret=client_secret,
        ),
        paginator=SinglePagePaginator(),
    )

    @dlt.resource(primary_key="id", write_disposition="replace")
    def workspaces() -> Generator[list[dict[str, Any]], None, None]:
        endpoint = "groups"
        yield client.get(endpoint).json()["value"]

    # We need to add the type hint for our custom column here as this is required for
    # dbt-osmosis to correctly generate the bronze properties file, and consequently,
    # for dbt-ibis to work.
    @dlt.transformer(
        data_from=workspaces, columns={"description": {"data_type": "text"}}
    )
    def workspaces_lineage(
        workspaces: list[Workspace],
    ) -> Generator[list[Workspace], None, None]:
        workspace_ids = [workspace["id"] for workspace in workspaces]
        request_lineage_endpoint = "admin/workspaces/getInfo"
        params = {
            "lineage": True,
            "datasourceDetails": True,
            "datasetSchema": True,
            "datasetExpressions": True,
            "getArtifactUsers": True,
        }
        body = {"workspaces": workspace_ids}

        # Request a workspace lineage scan and await scan completion.
        response = client.post(request_lineage_endpoint, params=params, json=body)
        scan_id = response.json()["id"]
        scan_status = None
        logger.info("Waiting for scan to complete...")
        while scan_status != "Succeeded":
            scan_status_endpoint = f"admin/workspaces/scanStatus/{scan_id}"
            scan_status = client.get(scan_status_endpoint).json()["status"]
            sleep(0.2)

        # Get the scan result.
        scan_result_endpoint = f"admin/workspaces/scanResult/{scan_id}"
        response = client.get(scan_result_endpoint)
        lineage = response.json()

        # Add "description" column if it doesn't exist.
        # This is required as the schema returned by PowerBI REST API is dynamic, which
        # can break everything downstream. For example, if a workspace description is
        # not set, instead of returning "description": "", PBI REST API simply omits the
        # "description" key.
        for i, workspace in enumerate(lineage["workspaces"]):
            if "description" not in workspace:
                lineage["workspaces"][i]["description"] = ""

        yield lineage

    return [workspaces_lineage]


if __name__ == "__main__":
    import json

    # for workspace in powerbi():
    # print(f"Workspace lineage:\n{workspace}")
    # logger.info(json.dumps(workspace, indent=4))
    with Path("powerbi_workspace_info.json").open("w", encoding="utf-8") as f:
        json.dump(list(powerbi()), f, indent=4)



from enum import Enum
from typing import Any

from pydantic import BaseModel, EmailStr


class ColumnType(str, Enum):
    CalculatedTableColumn = "CalculatedTableColumn"
    Calculated = "Calculated"
    Data = "Data"


class TableColumn(BaseModel):
    name: str
    dataType: str
    isHidden: bool
    columnType: ColumnType


class TableMeasure(BaseModel):
    name: str
    expression: str
    description: str | None = None
    isHidden: bool


class TableSource(BaseModel):
    expression: str


class Table(BaseModel):
    name: str
    isHidden: bool
    columns: list[TableColumn] = []
    measures: list[TableMeasure] = []
    source: list[TableSource] | None = None


class Dataset(BaseModel):
    id: str
    name: str
    tables: list[Table] = []
    description: str | None = None
    configuredBy: EmailStr | None = None
    configuredById: str | None = None
    directQueryRefreshSchedule: dict[str, Any] | None = None
    createdDate: str
    users: list[dict] | None = None


class DashboardTile(BaseModel):
    id: str
    title: str
    reportId: str
    datasetId: str


class Dashboard(BaseModel):
    id: str
    displayName: str
    tiles: list[DashboardTile]
    users: list[dict]
    tags: list[str]


class ReportUser(BaseModel):
    reportUserAccessRight: str
    emailAddress: EmailStr | None
    displayName: str
    identifier: EmailStr | str
    graphId: str
    principalType: str
    userType: str | None


class Report(BaseModel):
    id: str
    appId: str | None = None
    name: str
    datasetId: str
    description: str | None = None
    createdDateTime: str | None = None
    modifiedDateTime: str | None = None
    users: list[ReportUser]


class Workspace(BaseModel):
    id: str
    name: str
    type: str = "Workspace"
    state: str
    reports: list[Report] = []
    datasets: list[Dataset] = []
    dashboards: list[Dashboard] = []
    users: list[dict]


class WorkspaceInfo(BaseModel):
    workspaces: list[Workspace]
    datasourceInstances: list[dict] | None = None



import uuid

from lumaCLI.luma import app
from typer.testing import CliRunner

runner = CliRunner()


def test_status():
    fake_uuid = uuid.uuid4()
    result = runner.invoke(
        app,
        ["status", "--luma-url", "http://localhost:8000", "--ingestion-id", fake_uuid],
    )
    assert result.exit_code == 0



import pytest

from lumaCLI.luma import app
from typer.testing import CliRunner

runner = CliRunner()

METADATA_FIXTURE_NAMES = ["METADATA_DIR_V1_7", "METADATA_DIR_V1_9"]

@pytest.mark.parametrize("metadata_dir_fixture", METADATA_FIXTURE_NAMES)
def test_ingest(request, metadata_dir_fixture):

    metadata_dir = request.getfixturevalue(metadata_dir_fixture)

    result = runner.invoke(
        app,
        [
            "dbt",
            "ingest",
            "--metadata-dir",
            metadata_dir / "model_metadata",
            "--luma-url",
            "http://localhost:8000",
            "--dry-run",
        ],
    )
    assert result.exit_code == 0


@pytest.mark.parametrize("metadata_dir_fixture", METADATA_FIXTURE_NAMES)
def test_ingest_with_shorthand(request, metadata_dir_fixture):

    metadata_dir = request.getfixturevalue(metadata_dir_fixture)

    result = runner.invoke(
        app,
        [
            "dbt",
            "ingest",
            "-m",
            metadata_dir / "model_metadata",
            "-l",
            "http://localhost:8000",
            "-D",
        ],
    )
    assert result.exit_code == 0
    assert "Dry run mode" in result.stdout



import pytest
from typer.testing import CliRunner

from lumaCLI.luma import app


runner = CliRunner()


@pytest.mark.skip
def test_ingest(test_server, setup_db):
    """
    Test the 'ingest' command of the CLI application to ensure it successfully
    ingests data from a specified source into a PostgreSQL database.

    Args:
    - test_server: A fixture providing the test server URL.
    - setup_db: A fixture providing database connection details.

    The test invokes the CLI 'ingest' command with the necessary arguments
    and checks the exit code and output to verify successful execution.
    """
    result = runner.invoke(
        app,
        [
            "postgres",
            "ingest",
            "--luma-url",
            test_server,
            "--database",
            setup_db.info.dbname,
            "--host",
            setup_db.info.host,
            "--port",
            setup_db.info.port,
            "--username",
            setup_db.info.user,
            "--password",
            setup_db.info.password,
            "--no-config",
        ],
    )
    assert result.exit_code == 0
    assert "The request was successful!" in result.output
